"""Various html generating utility functions"""

import uuid

from mara_page import _


def highlight_syntax(code: str, language: str) -> [str]:
    """
    Renders a code snipped into a syntax-highlighted html fragment

    Args:
        code: The code to render
        language: See "short names" in http://pygments.org/docs/lexers/
    Returns:
        html markup
    """
    import pygments.formatters
    import pygments.lexers

    formatter = pygments.formatters.HtmlFormatter(nobackground=True, style='friendly')
    return [_.style[formatter.get_style_defs()],
            pygments.highlight(code, pygments.lexers.get_lexer_by_name(language, strip_all=True), formatter=formatter)]


def spinner() -> [str]:
    """
    Returns markup for an animated load spinner.

    The default version requires font-awesome to be installed

    Returns: html markup
    """

    return _.span(class_='fa fa-spinner fa-spin')[' ']


def spinner_js_function() -> [str]:
    """Creates a javascript function that returns the markup of `spinner` """
    return _.script["function spinner() { return '" + str(spinner()) + "' };"]


def asynchronous_content(url: str, div_id: str = None) -> [str]:
    """
    Creates a div whose content will be asynchronously replaced with the content retrieved from `url`.

    Requires the implementation of the javascript function `loadContentAsynchronously` that takes four arguments
    - the container div
    - the url to load
    - a localStorage key for storing the final height of the div
    - an optional javascript snippet that is called once the content is loaded

    Args:
        url: The url from which to retrieve the content
        div_id: The id of the container div

    Returns:
        Html markup of the container div
    """
    div_id = div_id or str(uuid.uuid1())
    return _.div(id=div_id)[
        spinner(),
        _.script["""
 
(function() {
    // immediately (even before the DOM is completely loaded) set the height of the div 
    // to the last content height (stored in local storage) to avoid height flickering  
    var divHeightKey = 'div-height--' + window.location.pathname + '--' + '""" + url + """';
    var divHeight = localStorage.getItem(divHeightKey);
    if (divHeight) {
        document.getElementById('""" + div_id + """').style.height = divHeight + 'px';    
    }
    
    document.addEventListener('DOMContentLoaded', function() {
        if (typeof loadContentAsynchronously == 'undefined') {
            console.error('Please implement function "loadContentAsynchronously"');
        } else {
            loadContentAsynchronously('""" + div_id + """', '""" + url + """', divHeightKey);
        }
    });
})();
"""]]



"""A flask response class that adds titles, action buttons and assets"""

import typing

import flask


from mara_page.bootstrap import ActionButton

class Response(flask.Response):
    def __init__(self,
                 html,
                 title: str,
                 action_buttons: typing.Optional[typing.List[ActionButton]] = None,
                 js_files: typing.Optional[typing.List[str]] = None,
                 css_files: typing.Optional[typing.List[str]] = None,
                 status: int = 200):
        """
        A rich html response with additional information for applying a page layout

        Args:
            html: The content of the page
            title: The title to be displayed in the page header
            action_buttons: Global actions for the page
            js_files: URL paths of js files to include
            css_files: URL paths of css files to include
            status_code: The http status code
        """
        self.title = title
        self.action_buttons = action_buttons or []
        self.js_files = js_files or []
        self.css_files = css_files or []
        super().__init__(html, status)



"""Functions for creating bootstrap 4 compatible markup"""

import uuid

from mara_page import _

class ActionButton:
    def __init__(self, action: str, label: str, title: str, icon: str):
        """
        A button that is displayed at the top of a page

        Args:
            action: An url or `javascript:foo()` function call
            label: The label of the button
            title: A help text
            icon: The icon of the button (from http://fontawesome.io/)
        """
        self.action = action
        self.title = title
        self.icon = icon
        self.label = label


def card(header_left='', header_right='', fixed_header_height: bool = True,
         body=[], sections=[], id: str = None, action_buttons: [str] = None):
    """
    Renders a bootstrap card `bootstrap_card`_ 
    
    Args:
        header_left: A header that is displayed at the top left of the card
        header_right: A header that is displayed at the top right of the card
        fixed_header_height: When true, then the header is restricted to 1 line
        body: Elements to be shown on the card
        sections: Parts of the card that are separated by an horizontal line
        id: An optional id for the outer dom element of the card
    Returns:
        The rendered card
        
    .. _bootstrap_card:
       https://v4-alpha.getbootstrap.com/components/card/     
    """
    return _.div(id=id or uuid.uuid1(), class_="card mara-card")[
        (_.div(class_='card-header' + (' fixed-header-height' if fixed_header_height else ''))[
             (_.div(class_='card-header-left')[header_left] if header_left else ''),
             (_.div(class_='card-header-right')[header_right] if header_right else '')]
         if header_left != '' or header_right != ''
         else ''),

        (_.div(class_='card-block')[_.div(class_='card-block-content')[body]] if body else ''),
        [_.div(class_='card-block card-section')[_.div(class_='card-block-content')[section]]
         for section in sections] or ''
    ]


def table(headers: [str], rows: [], id: str = None):
    """
    Renders a bootstrap table with some defaults applied
     
    Args:
        headers: The column headers (list of strings)
        rows: All table rows (rendered trs)
        id: An optional id for the table
    Returns:
        The rendered table
    """
    return _.table(class_='mara-table table table-hover table-condensed table-sm'
                          + (' mara-table-float-header' if headers else ''),
                   id=id or uuid.uuid1())[
        _.thead[_.tr[[_.th[header] for header in headers]]],
        _.tbody[rows]]


def button(url: str, label: str, title: str, icon: str, id: str = None):
    """
    Renders a bootstrap button
    Args:
        url: The action to perform
        label: The button label
        title: A help message
        icon: An icon from the `fontawesome`_ collection
        id: An id that is added to the element
    Returns:
        The rendered button

    .. _fontawesome:
       http://fontawesome.io/icons/
    """
    return _.a(class_='btn mara-button', href=url, title=title, id=id or uuid.uuid1())[
        _.span(class_='fa fa-' + icon)[''], ' ', label]



"""A minimal API for defining ACL protected resources and querying permissions"""

import functools
import typing

import flask


class AclResource:
    def __init__(self,
                 name: str,
                 rank: int = 1,
                 children: typing.Optional[typing.List['AclResource']] = None):
        """
        A resource that is protected by an acl

        Args:
            name: An identifier of the resource, must be unique among its parents
            rank: How to sort entries within siblings
            description: An optional help text
            children: A list of sub-entries
        """
        self.name = name
        self.rank = rank
        self.children = []
        self.parent = None
        if children:
            for child in children:
                self.add_child(child)

    def add_child(self, child: 'AclResource'):
        self.children.append(child)
        child.set_parent(self)

    def set_parent(self, parent: 'AclResource'):
        """Sets the parent of the entry"""
        self.parent = parent

    def __repr__(self):
        return '<AclResource "' + self.name + '">'


def current_user_has_permission(resource: AclResource) -> bool:
    """
    Whether the current user is allowed to access a specific resource.
    """
    return current_user_has_permissions([resource])[0][1]


def current_user_has_permissions(resources: [AclResource]) -> [[AclResource, bool]]:
    """
    Determines whether the currently logged in user has permissions for a list of resources.
    Implement actual behavior by patching the function.
    """
    return list(map(lambda resource: [resource, True], resources))


def current_user_email():
    """
    Returns the email address of the currently logged in user.
    Implement actual behavior by patching the function.
    """
    return 'guest@localhost'


def user_has_permission(email: str, resource: AclResource) -> bool:
    """
    Whether a user is allowed to access a specific resource.
    """
    return user_has_permissions(email, [resource])[0][1]


def user_has_permissions(email: str, resources: [AclResource]) -> [[AclResource, bool]]:
    """
    Determines whether a user has permissions for a list of resources.
    Implement actual behavior by patching the function.
    """
    return map(lambda resource: [resource, True], resources)


def require_permission(resource: AclResource, do_abort: bool = True,
                       abort_message="Sorry, but you don't have enough permissions to view this page.",
                       return_message="Not enough permissions.") \
        -> typing.Callable:
    """
    A decorator for protecting a resource by acl

    Args:
        resource: The resource for which user permissions are required
        do_abort: When true, a http exception is raised if the the user does not have permission
                      (useful when protecting whole pages).
                  When false, a small error message is returned (useful for ajax handlers).
        abort_message: The text of the "permission denied" http exception
        return_message: The text of the returned "permission denied" inline content

    Returns: The wrapped function
    """

    def decorator(f):
        def wrapper(*args, **kwargs):
            if not current_user_has_permission(resource):
                if do_abort:
                    flask.abort(403, abort_message)
                else:
                    return inline_permission_denied_message(return_message)

            else:
                return f(*args, **kwargs)

        functools.update_wrapper(wrapper, f)

        return wrapper

    return decorator


def inline_permission_denied_message(message="Not enough permissions"):
    """Returns a an inline html element that signals insufficient permissions"""
    return f'<span style="font-style:italic;color:#aaa"><span class="fa fa-lock"> </span> {message}</span>'



"""Definition of navigation entries"""

import typing


class NavigationEntry:
    def __init__(self,
                 label: str,
                 uri_fn: typing.Callable = lambda: '#',
                 icon: str = None,
                 rank: int = 1,
                 description: str = None,
                 visible: bool = True,
                 children: typing.Optional[typing.List['NavigationEntry']] = None):
        """
        A single entry of the navigation sidebar.

        Args:
            label: The text that is displayed on the entry
            uri: A function that returns the uri of the linked page (not used when entry has children)
                 This needs to a callable because the app context for url creation is not known yet
            icon: An optional icon of the link (from http://fontawesome.io/)
            rank: How to sort entries within siblings
            description: An optional help text
            visible: When False, then the entry is not shown
            children: A list of sub-entries
        """
        self.label = label
        self.uri_fn = uri_fn
        self.icon = icon
        self.rank = rank
        self.description = description
        self.children = []
        self.parent = None
        self.visible = visible
        if children:
            for child in children:
                self.add_child(child)

    def add_child(self, child: 'NavigationEntry'):
        self.children.append(child)
        child.set_parent(self)

    def set_parent(self, parent: 'NavigationEntry'):
        """Sets the parent of the entry"""
        self.parent = parent

    def __repr__(self):
        return '<NavigationEntry "' + self.label + '">'

    def __hash__(self) -> int:
        return hash((self.label, self.icon, self.description))

    def __eq__(self, o: 'NavigationEntry') -> bool:
        return self.label == o.label and self.icon == o.icon and self.description == o.description




from setuptools import setup, find_packages
import re

def get_long_description():
    with open('README.md') as f:
        return re.sub('!\[(.*?)\]\(docs/(.*?)\)', r'![\1](https://github.com/mara/mara-page/raw/master/docs/\2)', f.read())

setup(
    name='mara-page',
    version='1.6.0',

    description='Minimal API for defining pages of Flask-based backends independently from the actual backend infrastructure',

    long_description=get_long_description(),
    long_description_content_type='text/markdown',

    url = 'https://github.com/mara/mara-page',

    python_requires='>=3.6',

    install_requires=[
        'flask>=0.12',
        'pygments>=2.2.0',
    ],

    extras_require={
        'test': ['pytest', 'pytest_click'],
    },

    packages=find_packages(),

    setup_requires=['setuptools_scm'],
    include_package_data=True,

    author='Mara contributors',
    license='MIT',

    entry_points={},
)



"""Expressing Html / XML document fragments in pure python syntax.

Examples:
    >>> print(_.div)
    <div/>

    >>> print(_.div['Hello world!'])
    <div>Hello world!</div>

    >>> print(_.div[''])
    <div></div>

    >>> print(_.div(style='color:red')['Hello world!'])
    <div style="color:red">Hello world!</div>

    >>> print(_.div(style='color:red', class_='foo')['Hello ', _.span(style='color:green')['world'], '!'])
    <div style="color:red" class="foo">Hello <span style="color:green">world</span>!</div>

    >>> print(_.ul[[_.li['item ', str(i)] for i in [1, 2, 3]]])
    <ul><li>item 1</li><li>item 2</li><li>item 3</li></ul>


It is very much inspired by `CL-WHO`_ for Common Lisp and `Hiccup`_ for Clojure.

For Python, there is also
 - `pyhiccup`_ (unfortunately GPL3 licenced)
 - `stan`_ (part of a web framework that is too big to have as a dependency)
 - `lmxl builder`_ (hidden in a huge XML library)

.. _CL-WHO:
   http://weitz.de/cl-who/

   _Hiccup:
   https://github.com/weavejester/hiccup

   _pyhiccup:
   https://pypi.python.org/pypi/pyhiccup

   _stan:
   https://github.com/twisted/nevow/blob/master/nevow/stan.py

   _lxml builder:
   http://lxml.de/api/lxml.html.builder-module.html

"""
import warnings


class XMLElement():
    """Representation of a html / xml element with attributes and children"""

    def __init__(self, tag_name):
        self.tag_name = tag_name
        self.attributes = {}
        self.children = []

    def __call__(self, **kwargs):
        """
        Adds attributes to the element. When the desired attribute name is a reserved
        python keyword, then postfix it with '_'
        """
        for attribute_name, value in kwargs.items():
            if attribute_name[0] == '_':
                warnings.warn('Prefixing python keywords is deprecated, please postfix such names', FutureWarning,
                              stacklevel=2)
                attribute_name = attribute_name[1:]
            if attribute_name[-1] == '_':
                attribute_name = attribute_name[:-1]
            self.attributes[attribute_name] = str(value)
        return self

    def __getitem__(self, *children):
        """Adds children to the element"""
        self.children = children
        return self

    def __str__(self):
        """Renders the element and it's children"""
        return ''.join(render(self))


def render(x) -> [str]:
    """Streams an expression into a list of strings"""
    if isinstance(x, str):
        yield x
    elif isinstance(x, bytes):
        yield x.decode('utf-8')
    elif hasattr(x, '__iter__'):
        for e in x:
            yield from render(e)
    elif isinstance(x, XMLElement):
        yield f'<{x.tag_name}'
        if x.attributes:
            for attribute, value in x.attributes.items():
                yield f' {attribute}="{value}"'
        if x.children:
            yield '>'
            for child in x.children:
                yield from render(child)
            yield f'</{x.tag_name}>'
        else:
            yield '/>'
    else:
        yield repr(x)


class XMLElementFactory():
    """Creates XML elements using the '.' operator"""

    def __init__(self):
        pass

    def __getattribute__(self, name) -> XMLElement:
        return XMLElement(name)


_ = XMLElementFactory()
"""Convenience shortcut for writing ``_.div['bla']`` instead of ``XMLElement('div')['bla']``"""
